unit FastStrm;
{
  A stream class for Fast Sector-Aligned disk write. Good for use with
  Video Streams.
  This is a write-only class that flushes the stream on Free/Destroy.

  Note that Seek has a very unstandard behavior, so don't rely on it !!
  Also read the note in Destroy, where I wanted the indicted behavior
  to keep a fixed-length file on disk.

  Version 1.3, some problems with Flush corrected.

  Freeware (c) by Dieter Menne, Biomed Software, Tbingen
  Compuserve 100016,2125
  Internet 100016.2125@compuserve.com
}

interface

uses
  Windows, Messages, SysUtils, Classes;

type
  TFastWriteStream = class(TStream)
    private
      FPosition: LongInt;
      FSectorSize,
      FSectors,
      FSize,
      FHandle,
      FWriteToIndex: Integer;
      Buf: PChar;

    public
      constructor Create(const FileName:string; MinBufferSize: DWORD);
      destructor Destroy; override;
      function Write(const Buffer; Count: Longint): Longint; override;
      function Seek(Offset: Longint; Origin: Word): Longint; override;
      procedure ReserveMB(Reserve: integer);
      property Position: longint read FPosition;
    end;

  function GetSectorSize(Disk: string): integer;

implementation

uses Consts,AuxOut;

{ ----------------------------------- GetSectorSize ---------------------------}
function GetSectorSize(Disk: string): integer;
var
  Drive: integer;
  RootPath: array[0..4] of Char;
  RootPtr: PChar;
  SectorsPerCluster,
  BytesPerSector,
  FreeClusters,
  TotalClusters: Integer;
begin
  Drive:=0; // Default Drive
  Disk:=UpperCase(Disk);
  if (Length(Disk) = 2) and (Disk[2]=':') then
   Drive:=ord(Disk[1])-ord('A')+1;  // A is drive 1
  RootPtr := nil;
  if Drive > 0 then
  begin
    StrCopy(RootPath, 'A:\');
    RootPath[0] := Char(Drive + $40);
    RootPtr := RootPath;
  end;
  if GetDiskFreeSpace(RootPtr, SectorsPerCluster, BytesPerSector,
    FreeClusters, TotalClusters) then
    Result := BytesPerSector
  else Result := -1;
end;

{ ------------------------------- Create --------------------------------------}
constructor TFastWriteStream.Create(const FileName: string;
                                        MinBufferSize: DWORD);
begin

  // Find out the Sector aligned Buffer size larger than MinBufferSize
  FSectorSize:=GetSectorSize(FileName);
  FSectors:=(MinBufferSize div FSectorSize+1);
  FSize:=FSectorSize*FSectors;

  // Note the FILE_FLAG_NOBUFFERING
  FHandle:=CreateFile(PChar(FileName),GENERIC_WRITE,0,nil,OPEN_ALWAYS,
                      FILE_ATTRIBUTE_NORMAL or FILE_FLAG_NO_BUFFERING,0);
  if FHandle < 0 then
      raise EFCreateError.CreateResFmt(SFCreateError, [FileName]);
  // The main trick: VirtualAlloc aligns at segment offset in memory
  // This memory is automatically set to zero
  Buf:=VirtualAlloc(nil,FSize,MEM_COMMIT,PAGE_READWRITE);
  FWriteToIndex:=0;
  if Buf = nil then
  begin
    CloseHandle(FHandle);
    raise EFCreateError.CreateResFmt(SFCreateError, [FileName]);
  end;
end;

{ ------------------------------ Destroy --------------------------------------}
destructor TFastWriteStream.Destroy;
var WriteAligned,Written: DWORD;
begin
  { write remaining Buffer }
  if FWriteToIndex > 0 then
  begin
    WriteAligned:=((FWriteToIndex div FSectorSize)+1)*FSectorSize;
    // We write too much..
    WriteFile(FHandle,Buf^,WriteAligned,Written,nil);

    // So we must truncate, but in my application I don't want this
    // to keep the large file open
    //SetFilePointer(FHandle,FPosition,nil,FILE_BEGIN);
    //SetEndOfFile(FHandle);
  end;
  VirtualFree(Buf,0,MEM_RELEASE);
  CloseHandle(FHandle);
  inherited Destroy;
end;

{ ------------------------------ Write ----------------------------------------}
function TFastWriteStream.Write(const Buffer; Count: Longint): Longint;
var FreeBytes: integer;
    Success: boolean;
    Written: DWORD;
begin
  { It would be easy to extend this to the case where the reserved buffer
  is smaller than the requested Block write, but it is anyway faster if
  you really know the largest size }
  if Count > FSize then
    raise Exception.Create('FastWriteStream: Allocated buffer to small');

  Result:=0;  // Default: write error
  FreeBytes:=FSize-FWriteToIndex;
  if FreeBytes > Count then  // Buffer not yet full, just copy all to buffer
  begin
    CopyMemory(@Buf[FWriteToIndex],@Buffer,Count);
    FWriteToIndex:=FWriteToIndex+Count;
  end
  else   // Buffer full, copy, write and move overhang
  begin
    CopyMemory(@Buf[FWriteToIndex], @Buffer, FreeBytes);
    Success:=WriteFile(FHandle,Buf^,FSize,Written,nil);
    if Not Success then exit; // Report write error
    FWriteToIndex:=Count-FreeBytes;
    if FWriteToIndex <> 0 then
      CopyMemory(@Buf[0],Pchar(@Buffer)+FreeBytes,FWriteToIndex);
  end;
  Inc(FPosition,Count);
  Result:=Count; // Report success to WriteBuffer where exceptions are handled
end;

{ ------------------------------ Seek ------------------------------------------}
function TFastWriteStream.Seek(Offset: Longint; Origin: Word): Longint;
var WriteAligned,Written: integer;
begin
  Result:=0;
  if (Offset <> 0) and (Origin <> soFromBeginning) then
    raise Exception.Create
    ('FastWriteStreamSeek: Only Offset=0, soFromBeginning permitted');
  // Flush current buffer
  if FWriteToIndex > 0 then
  begin
    WriteAligned:=((FWriteToIndex div FSectorSize)+1)*FSectorSize;
    WriteFile(FHandle,Buf^,WriteAligned,Written,nil);
    FWriteToIndex:=0;
  end;
  SetFilePointer(FHandle,0,nil,FILE_BEGIN);
  FPosition:=0;
end;

{ ------------------------------ ReserveMB ------------------------------------}
procedure TFastWriteStream.ReserveMB(Reserve:integer);
begin
  if Position <> 0 then
    raise Exception.Create
    ('FastWriteStreamSeek: ReserveMB must be called before writing');
  SetFilePointer(FHandle,Reserve*1024*1024,nil,FILE_BEGIN);
  SetEndOfFile(FHandle);
  SetFilePointer(FHandle,0,nil,FILE_BEGIN);
end;


end.

